@@ -5,6 +5,7 @@
 #include <functional>
 #include <memory>
 #include <iostream> // for debug
+#include <cstring>
 
 
 #include <stdio.h>
@@ -18,7 +19,7 @@
 
 namespace panda { namespace backtrace {
 
-static StackframePtr dl_produce(const void*);
+static BacktraceBackendSP dl_produce(const Backtrace& raw_traces);
 static BacktraceProducer dl_producer(dl_produce);
 
 static DwarfInfoMap load_dwarf_info(const SharedObjectMap& so_map) {
@@ -50,19 +51,30 @@ static DwarfInfoMap load_dwarf_info(const SharedObjectMap& so_map) {
     return result;
 }
 
+struct DwarfBackend: BacktraceBackend {
+    const Backtrace& raw_traces;
+    DwarfInfoMap info_map;
 
-StackframePtr dl_produce(const void* frame_ptr) {
-    SharedObjectMap so_map;
-    std::vector<StackframePtr> frames;
-    gather_info(so_map);
-    auto map = load_dwarf_info(so_map);
-    auto ip_addr = reinterpret_cast<std::uint64_t>(frame_ptr);
-    for(auto& it: map) {
-        //std::cout << "resolving " << it.first << "\n";
-        auto frame = it.second->resolve(ip_addr);
-        if (frame) return frame;
+    DwarfBackend(const Backtrace& raw_traces_) noexcept: raw_traces{raw_traces_} {
+        SharedObjectMap so_map;
+        gather_info(so_map);
+        info_map = load_dwarf_info(so_map);
     }
-    return StackframePtr();
+
+    bool produce_frame(StackFrames& frames, size_t i) override {
+        auto frame_ptr = raw_traces.buffer.at(i);
+        auto ip_addr = reinterpret_cast<std::uint64_t>(frame_ptr);
+        for(auto& it: info_map) {
+            //std::cout << "resolving " << it.first << "\n";
+            if (it.second->resolve(ip_addr, frames)) return true;
+        }
+        return false;
+    }
+
+};
+
+BacktraceBackendSP dl_produce(const Backtrace& raw_traces) {
+    return new DwarfBackend(raw_traces);
 }
 
 void install_backtracer() {
@@ -85,7 +97,7 @@ bool DwarfInfo::load(file_guard_t&& guard_) noexcept {
     guard = std::move(guard_);
     Dwarf_Error error;
     for(int cu_number = 0;;++cu_number) {
-        auto cu = std::make_unique<dwarf::CU>(debug, cu_number);
+        auto cu = dwarf::CUSP(new dwarf::CU(debug, cu_number));
 
         auto res = dwarf_next_cu_header_d(debug, true, &cu->header_length,
              &cu->version_stamp, &cu->abbrev_offset,
@@ -97,7 +109,7 @@ bool DwarfInfo::load(file_guard_t&& guard_) noexcept {
 
         /* The CU will have a single sibling, a cu_die. */
         Dwarf_Die cu_die = nullptr;
-        res = dwarf_siblingof_b(debug, nullptr, true,&cu_die, &error);
+        res = dwarf_siblingof_b(debug, nullptr, true, &cu_die, &error);
         if (res != DW_DLV_OK) { break; }
 
         dwarf::DieHolder cu_die_holder(cu_die, debug, nullptr);
@@ -108,63 +120,77 @@ bool DwarfInfo::load(file_guard_t&& guard_) noexcept {
 }
 
 
-StackframePtr DwarfInfo::resolve(std::uint64_t ip) noexcept {
-    if (ip < so_info.begin || ip >= so_info.end) { return StackframePtr(); }
+bool DwarfInfo::resolve(std::uint64_t ip, StackFrames &frames) noexcept {
+    if (ip < so_info.begin || ip >= so_info.end) { return false; }
 
     auto offset = so_info.get_offset(ip);
-    ///std::cout << "resolving " << std::hex << ip << "/" << offset << " from " <<  so_info.name << ", CUs: " << CUs.size() << "\n";
+    /// std::cout << "resolving " << std::hex << ip << "/" << offset << " from " <<  so_info.name << ", CUs: " << CUs.size() << "\n";
 
     for(auto it = CUs.begin(); it != CUs.end(); ++it){
         //if (r.is_complete()) std::cout << "hit\n";
         auto& cu = *it;
         auto r = cu->resolve(offset);
-        if (r.is_complete()) { return r.get_frame(ip, so_info); }
+        if (r.is_complete()) { return r.get_frames(ip, so_info, frames); }
     }
 
     // just fall-back to .so, address & offset
-    auto frame = StackframePtr(new Stackframe());
+    auto frame = StackframeSP(new Stackframe());
     frame->library = so_info.name;
     frame->address = ip;
     frame->offset  = offset;
-    return frame;
+    frames.emplace_back(std::move(frame));
+    return true;
 }
 
 namespace dwarf {
 
 LookupResult::LookupResult(LookupResult&& other) {
     cu = std::move(other.cu);
+    root = std::move(other.root);
     subprogram = std::move(other.subprogram);
     offset = std::move(other.offset);
 }
 bool LookupResult::is_complete() noexcept { return cu && subprogram; }
 
-StackframePtr LookupResult::get_frame(std::uint64_t ip, const SharedObjectInfo& so) noexcept {
-    auto frame = StackframePtr(new Stackframe());
+bool LookupResult::get_frames(std::uint64_t ip, const SharedObjectInfo& so, StackFrames &frames) noexcept {
     //if (!is_complete()) { return frame; }
 
-    //std::cout << "found die: " << (void*) die.die << ", h:" << addr->high << ", l: " << addr->low << ", o:" << offset << "\n";
+    auto push_frame = [&](const auto& details) {
+        auto frame = StackframeSP(new Stackframe());
+        frame->address = ip;
+        frame->offset = offset;
+        frame->library = so.name;
 
-    frame->address = ip;
-    frame->offset = offset;
-    frame->library = so.name;
+        if (details.name)    frame->name    = details.name;
+        if (details.line_no) frame->line_no = details.line_no;
+        if (details.source)  frame->file    = details.source;
+
+        frames.emplace_back(std::move(frame));
+    };
 
     if (subprogram) {
-        auto details = subprogram->refine_fn(*this);
-        frame->name = details.name;
-        frame->line_no = details.line_no;
-    }
+        auto location =  subprogram->refine_location(offset);
+        auto details = location->refine_fn(*this);
+        push_frame(details);
 
-    if (cu) {
-        auto details = cu->refine_cu();
-        frame->file = details.name;
+        // printf("refined: %s at %lu, offset: %lu(%lx)\n", details.name.c_str(), details.line_no, offset, offset);
+        while(!location->context.empty()) {
+            auto outer = location->context.back();
+            location->context.pop_back();
+            auto details = outer->refine_fn(*this);
+            push_frame(details);
+            //printf("refined (outer): %s at %u\n", details.name.c_str(), details.line_no);
+        }
+    } else {
+        push_frame(FunctionDetails());
     }
 
     ///std::cout << frame->name << " at " << frame->file << ":" << frame->line_no << ", o:" << frame->offset << "\n";
-    return frame;
+    return true;
 }
 
 
-DieRC::DieRC(Dwarf_Die die_, Dwarf_Debug debug_, panda::iptr<DieRC> parent_): die{die_}, debug{debug_}, parent{parent_} {}
+DieRC::DieRC(Dwarf_Die die_, Dwarf_Debug debug_, DieSP parent_): die{die_}, debug{debug_}, parent{parent_} {}
 DieRC::~DieRC() {
     dwarf_dealloc(debug, die,DW_DLA_DIE);
 }
@@ -185,7 +211,7 @@ Dwarf_Die DieRC::resolve_ref(Dwarf_Die source, Dwarf_Half attr) noexcept {
     return nullptr;
 }
 
-panda::iptr<DieRC> DieRC::discover(Dwarf_Die target) noexcept {
+DieSP DieRC::discover(Dwarf_Die target) noexcept {
     auto p = parent;
     while (p->parent) { p = p->parent; }
 
@@ -207,7 +233,28 @@ panda::iptr<DieRC> DieRC::discover(Dwarf_Die target) noexcept {
     std::abort();
 }
 
-panda::iptr<DieRC> DieRC::discover(Dwarf_Off target_offset, DieHolder& node) noexcept {
+template<typename CheckFN>
+DieSP traverse_sibling_or_child(DieHolder& node, CheckFN&& fn){
+    Dwarf_Error error;
+    Dwarf_Die child_die;
+    int res = dwarf_siblingof_b(node.debug, node.die, true, &child_die, &error);
+    if (res == DW_DLV_OK) {
+        DieHolder child(child_die, node.debug, node.parent);
+        auto found = fn(child);
+        if (found) { return found; }
+    }
+
+    // in-depth: check for child
+    res = dwarf_child(node.die, &child_die, &error);
+    if (res == DW_DLV_OK) {
+        DieHolder child(child_die, node.debug, &node);
+        auto found = fn(child);
+        if (found) { return found; }
+    }
+    return DieSP();
+}
+
+DieSP DieRC::discover(Dwarf_Off target_offset, DieHolder& node) noexcept {
     Dwarf_Error error;
     Dwarf_Off off;
     int res;
@@ -215,26 +262,9 @@ panda::iptr<DieRC> DieRC::discover(Dwarf_Off target_offset, DieHolder& node) noe
     res = dwarf_dieoffset(node.die, &off, &error);
     assert(res == DW_DLV_OK);
     if (off == target_offset) { return node.detach(); }
-    if (off > target_offset)  { return panda::iptr<DieRC>(); } /* do not lookup for fail branch */
+    if (off > target_offset)  { return DieSP(); } /* do not lookup for fail branch */
 
-    // in-breadth: check for siblings
-    Dwarf_Die child_die;
-    res = dwarf_siblingof_b(debug, node.die, true, &child_die, &error);
-    if (res == DW_DLV_OK) {
-        DieHolder child(child_die, debug, node.parent);
-        auto found = discover(target_offset, child);
-        if (found) { return found; }
-    }
-
-    // in-depth: check for child
-    res = dwarf_child(node.die, &child_die, &error);
-    if (res == DW_DLV_OK) {
-        DieHolder child(child_die, debug, &node);
-        auto found = discover(target_offset, child);
-        if (found) { return found; }
-    }
-
-    return panda::iptr<DieRC>();
+    return traverse_sibling_or_child(node, [&](DieHolder& child){ return discover(target_offset, child);  });
 }
 
 
@@ -243,18 +273,42 @@ void DieRC::refine_fn_name(Dwarf_Die it, FunctionDetails& details) noexcept {
         Dwarf_Error error;
         Dwarf_Attribute attr_name;
         auto res = dwarf_attr(it, DW_AT_name, &attr_name, &error);
+
         if (res == DW_DLV_OK) {
             iptr<DieRC> node = (it == die) ? iptr<DieRC>(this) : discover(it);
-            details.name = node->gather_fqn();
+            auto fqn = node->gather_fqn();
+            details.name = fqn.full_name;
+            details.name_die = fqn.source_die;
+            return;
         }
+
+        auto die_spec = resolve_ref(it, DW_AT_specification);
+        if (die_spec) return refine_fn_spec(die_spec, details);
+
+        auto die_ao = resolve_ref(it, DW_AT_abstract_origin);
+        if (die_ao) return refine_fn_name(die_ao, details);
     }
 }
 
-string DieRC::gather_fqn() noexcept {
+DieRC::FQN DieRC::gather_fqn() noexcept {
     Dwarf_Error error;
+    DieSP source_die;
+
+    auto try_record_source = [&](DieSP it) mutable {
+        Dwarf_Bool has_source;
+        if (!source_die) {
+            int res = dwarf_hasattr(it->die, DW_AT_decl_file, &has_source, &error);
+            if (res == DW_DLV_OK && has_source) {
+                source_die = it;
+            }
+        }
+    };
+
+
     char* name = nullptr;
     auto res = dwarf_diename(die, &name, &error);
     assert(res == DW_DLV_OK);
+    try_record_source(DieSP(this));
 
     string r(name);
 
@@ -271,30 +325,29 @@ string DieRC::gather_fqn() noexcept {
                 dwarf_formstring(attr_name, &prefix, &error);
                 assert(res == DW_DLV_OK);
                 r = string(prefix) + "::" + r;
+                try_record_source(p);
             }
         }
         p = p->parent;
     }
 
-    return r;
+    return FQN{r, source_die};
 }
 
-void DieRC::refine_fn_line(LookupResult& lr, FunctionDetails& details) noexcept {
-    /* currently it detects lines only in the current CU (compilation unit */
+void DieRC::refine_fn_line(Dwarf_Die die, std::uint64_t offset, FunctionDetails& details) noexcept {
+    /* currently it detects lines only in the current CU (compilation unit) */
     using LineContextHolder = std::unique_ptr<Dwarf_Line_Context, std::function<void(Dwarf_Line_Context *)>>;
-    auto& cu = lr.cu;
-    if (!cu) { return; }
 
     Dwarf_Error error;
     char* cu_name_raw;
-    auto res = dwarf_die_text(cu->die, DW_AT_name, &cu_name_raw, &error);
+    auto res = dwarf_die_text(die, DW_AT_name, &cu_name_raw, &error);
     if (res != DW_DLV_OK) { return; }
     string cu_name(cu_name_raw);
 
     Dwarf_Unsigned line_version;
     Dwarf_Small table_type;
     Dwarf_Line_Context line_context;
-    res = dwarf_srclines_b(cu->die, &line_version, &table_type,&line_context,&error);
+    res = dwarf_srclines_b(die, &line_version, &table_type,&line_context,&error);
     if (res != DW_DLV_OK) { return; }
     LineContextHolder line_context_guard(&line_context, [](auto it){ dwarf_srclines_dealloc_b(*it); });
 
@@ -353,8 +406,8 @@ void DieRC::refine_fn_line(LookupResult& lr, FunctionDetails& details) noexcept
         if (res != DW_DLV_OK) { return; }
         if (file_index != static_cast<Dwarf_Unsigned>(cu_index)) { continue; }
 
-        if (lineaddr >= lr.offset) { found = true; break;  }
-        else                       { prev_lineno = lineno; }
+        if (lineaddr >=  offset) { found = true; break;  }
+        else                     { prev_lineno = lineno; }
     }
 
     if (found) {
@@ -373,7 +426,7 @@ void DieRC::refine_fn_line_fallback(Dwarf_Die it, FunctionDetails& details) noex
         if (res == DW_DLV_OK) {
             Dwarf_Unsigned line;
             res = dwarf_formudata(attr_line, &line, &error);
-            if (res == DW_DLV_OK) { details.line_no = line; }
+            if (res == DW_DLV_OK) { details.line_no = line + 1; }
         }
     }
 }
@@ -382,30 +435,37 @@ FunctionDetails DieRC::refine_fn(LookupResult& lr) noexcept {
     FunctionDetails r;
 
     refine_fn_name(die, r);
-    refine_fn_line(lr, r);
-    refine_fn_line_fallback(die, r);
+    refine_fn_line(lr.cu->die, lr.offset, r);
+    if (!r.line_no && r.name_die) refine_fn_line_fallback(r.name_die->die, r);
 
-    if (!r.line_no || !r.name) {
-        auto die_spec = resolve_ref(die, DW_AT_specification);
-        if (die_spec) {
-            refine_fn_spec(die_spec, r);
-        }
-    }
-
-    if (!r.line_no || !r.name) {
-        auto die_ao = resolve_ref(die, DW_AT_abstract_origin);
-        if (die_ao) {
-            refine_fn_ao(die_ao, r);
-            if (die_ao) { refine_fn_ao(die_ao, r); }
-        }
-    }
+    if (r.name_die) refine_fn_source(r.name_die->die, r, *lr.root);
+    //printf("n = %s\n", r.name ? r.name.c_str() : "n/a");
 
     return r;
 }
 
+void DieRC::refine_fn_source(Dwarf_Die it, FunctionDetails& details, CU& cu) noexcept {
+    if (!details.source) {
+        Dwarf_Error error;
+        Dwarf_Attribute attr_file;
+        auto res = dwarf_attr(it, DW_AT_decl_file, &attr_file, &error);
+        if (res == DW_DLV_OK) {
+            Dwarf_Unsigned file_index;
+            res = dwarf_formudata(attr_file, &file_index, &error);
+            if (res == DW_DLV_OK && file_index) {
+                details.source = cu.get_source(file_index);
+            }
+        }
+    }
+}
+
+
 void DieRC::refine_fn_ao(Dwarf_Die abstract_origin, FunctionDetails& details) noexcept {
-    auto die_spec = resolve_ref(abstract_origin, DW_AT_specification);
-    if (die_spec) { refine_fn_spec(die_spec, details); }
+    refine_fn_name(abstract_origin, details);
+    if (!details.name) {
+        auto die_spec = resolve_ref(abstract_origin, DW_AT_specification);
+        if (die_spec) { refine_fn_spec(die_spec, details); }
+    }
 }
 
 
@@ -414,17 +474,78 @@ void DieRC::refine_fn_spec(Dwarf_Die specification, FunctionDetails& details) no
     refine_fn_line_fallback(specification, details);
 }
 
+template<typename CheckFN>
+Scan traverse(DieHolder& die, CheckFN&& fn) noexcept {
+    //std::cout << "resolving die: " << (void*) die.die << "\n";
 
-CUDetails DieRC::refine_cu() noexcept {
-    CUDetails r;
+    auto er = fn(die);
+    if (er == Scan::found) return er;
+
+    Dwarf_Die child_die = nullptr;
     Dwarf_Error error;
-    char* name = nullptr;
-    auto res = dwarf_die_text(die, DW_AT_name, &name, &error);
-    if (res == DW_DLV_OK && name) { r.name = name; }
-    return r;
+    int res;
+
+    // in-breadth: check for siblings
+    res = dwarf_siblingof_b(die.debug, die.die, true, &child_die, &error);
+    if (res == DW_DLV_OK) {
+        DieHolder child(child_die, die.debug, die.parent);
+        auto sr = traverse(child, fn);
+        if (sr == Scan::found) return sr;
+    } else if (res == DW_DLV_NO_ENTRY) {
+        /* ignore */
+    } else {
+        return Scan::dead_end;
+    }
+
+    // in-depth: check for child
+    if (er != Scan::dead_end) {
+        res = dwarf_child(die.die, &child_die, &error);
+        if(res == DW_DLV_OK) {
+            DieHolder child(child_die, die.debug, &die);
+            auto sr = traverse(child, fn);
+            if (sr == Scan::found) return sr;
+        }
+    }
+
+    return Scan::dead_end;
 }
 
-DieHolder::DieHolder(panda::iptr<DieRC> owner_):die{owner_->die}, debug{owner_->debug}, parent{nullptr}, owner{owner_} {
+DieSP DieRC::refine_location(uint64_t offset) noexcept {
+    DieCollection context{{DieSP(this)}};
+    DieHolder root(context.front());
+
+    Dwarf_Die child_die = nullptr;
+    Dwarf_Error error;
+    int res = dwarf_child(die, &child_die, &error);
+    if (res == DW_DLV_OK) {
+        DieHolder child(child_die, debug, &root);
+        traverse(child, [&](DieHolder& node) mutable {
+            Dwarf_Error error;
+            Dwarf_Half tag = 0;
+
+            res = dwarf_tag(node.die, &tag, &error);
+            if (res != DW_DLV_OK) { return Scan::dead_end; }
+
+            if( tag == DW_TAG_subprogram ||
+                tag == DW_TAG_inlined_subroutine) {
+                switch(node.contains(offset)) {
+                    case Match::no:  return Scan::dead_end;
+                    case Match::yes: context.push_back(node.detach()); break;
+                    default: break;
+                }
+            }
+            /* scan everything */
+            return Scan::not_found;
+        });
+    }
+
+    auto best = context.back();
+    context.pop_back();
+    best->context = std::move(context);
+    return best;
+}
+
+DieHolder::DieHolder(DieSP owner_):die{owner_->die}, debug{owner_->debug}, parent{nullptr}, owner{owner_} {
     assert(!owner || owner->die);
 }
 DieHolder::DieHolder(Dwarf_Die die_, Dwarf_Debug debug_, DieHolder* parent_): die{die_}, debug{debug_}, parent{parent_}{
@@ -434,10 +555,10 @@ DieHolder::~DieHolder() {
     if (!owner) { dwarf_dealloc(debug, die,DW_DLA_DIE); }
 }
 
-panda::iptr<DieRC> DieHolder::detach(){
+DieSP DieHolder::detach(){
     if (!owner) {
-        panda::iptr<DieRC> parent_ptr(parent ? parent->detach() : nullptr);
-        owner = panda::iptr<DieRC>(new DieRC(die, debug, parent_ptr));
+        DieSP parent_ptr(parent ? parent->detach() : nullptr);
+        owner = DieSP(new DieRC(die, debug, parent_ptr));
     }
     return owner;
 }
@@ -510,61 +631,36 @@ Match DieHolder::contains(std::uint64_t offset) noexcept {
 
 
 CU::CU(Dwarf_Debug debug_, int number_): debug{debug_}, number{number_} {
-    memset(&signature,0, sizeof(signature));
+    std::memset(&signature, 0, sizeof(signature));
+}
+
+CU::~CU() {
+    if (sources) {
+        for(size_t i = 0; i < static_cast<size_t>(sources_count); ++i) {
+             dwarf_dealloc(debug, sources[i], DW_DLA_STRING);
+        }
+        dwarf_dealloc(debug, sources, DW_DLA_LIST);
+    }
 }
 
 LookupResult CU::resolve(std::uint64_t offset) noexcept {
     assert(cu_die);
-    LookupResult lr;
+    LookupResult lr(*this);
     DieHolder dh(cu_die);
     resolve(offset, dh, lr);
     return lr;
 }
 
-bool CU::resolve(std::uint64_t offset, DieHolder& die, LookupResult& lr) noexcept {
-    //std::cout << "resolving die: " << (void*) die.die << "\n";
-
-    auto er = examine(offset, die, lr);
-    if (er || lr.is_complete()) { return er; }
-
-    Dwarf_Die child_die = nullptr;
-    Dwarf_Error error;
-    int res;
-
-    // in-breadth: check for siblings
-    res = dwarf_siblingof_b(debug, die.die, true, &child_die, &error);
-    if (res == DW_DLV_OK) {
-        DieHolder child(child_die, debug, die.parent);
-        resolve(offset, child, lr);
-        if (lr.is_complete()) { return true; }
-    } else if (res == DW_DLV_NO_ENTRY) {
-        /* ignore */
-    } else {
-        return true;
-    }
-
-    // in-depth: check for child
-    res = dwarf_child(die.die, &child_die, &error);
-    if(res == DW_DLV_OK) {
-        DieHolder child(child_die, debug, &die);
-        resolve(offset, child, lr);
-        if (lr.is_complete()) { return true; }
-    } else if (res == DW_DLV_NO_ENTRY) {
-        /* ignore */
-        return true;
-    } else {
-        return true;
-    }
-
-    return true;
+Scan CU::resolve(std::uint64_t offset, DieHolder& die, LookupResult& lr) noexcept {
+    return traverse(die, [&](DieHolder& node){ return examine(offset, node, lr); });
 }
 
-bool CU::examine(std::uint64_t offset, DieHolder &die, LookupResult& lr) noexcept {
+Scan CU::examine(std::uint64_t offset, DieHolder &die, LookupResult& lr) noexcept {
     Dwarf_Error error;
     Dwarf_Half tag = 0;
     assert(die.die);
     auto res = dwarf_tag(die.die, &tag, &error);
-    if (res != DW_DLV_OK) { return true; }
+    if (res != DW_DLV_OK) { return Scan::dead_end; }
 
     if( tag == DW_TAG_subprogram ||
         tag == DW_TAG_inlined_subroutine) {
@@ -572,25 +668,37 @@ bool CU::examine(std::uint64_t offset, DieHolder &die, LookupResult& lr) noexcep
         case Match::yes: {
             lr.subprogram = die.detach();
             lr.offset = offset;
-            return lr.is_complete();
+            return Scan::found;
         }
-        case Match::unknown: return false;
-        case Match::no:      return false;
+        default: return Scan::not_found;
         }
     }
     else if(tag == DW_TAG_compile_unit) {
         switch (die.contains(offset)) {
         case Match::yes: {
             lr.cu = die.detach();
-            return lr.is_complete();
+            return lr.is_complete() ? Scan::found : Scan::not_found;
         }
-        case Match::unknown: return false;
-        case Match::no:      return true;
+        case Match::unknown: return Scan::not_found;
+        case Match::no:      return Scan::dead_end;
         }
     }
 
     /* keep scaning */
-    return false;
+    return Scan::not_found;
 }
 
+string CU::get_source(size_t index) noexcept {
+    if (!sources_count) {
+        auto res = dwarf_srcfiles(cu_die->die, &sources, &sources_count, nullptr);
+        if (res != DW_DLV_OK) { sources_count = -1; }
+    }
+    if (sources_count > 0 && index < static_cast<size_t>(sources_count)) {
+        /* "subtract 1 to index into srcfiles", see dwarf_line.c */
+        return string(sources[index - 1]);
+    }
+    return string{};
+}
+
+
 }}}
